#include <dshow.h>
#include <winscard.h>

#include <stdio.h>

#include "hduscard.h"

static IBaseFilter *find_dev(
  REFCLSID category,BSTR varFriendlyName)
{
  HRESULT hr;
  ICreateDevEnum *pSysDevEnum;
  IEnumMoniker *pEnumCat;
  IMoniker *pMoniker;
  IPropertyBag *pPropBag;
  IBaseFilter *pFilter=NULL;
  ULONG cFetched;
  VARIANT varName;

  hr = CoCreateInstance(
    &CLSID_SystemDeviceEnum,NULL,CLSCTX_INPROC,
    &IID_ICreateDevEnum, (void **)&pSysDevEnum);
  if (hr != S_OK) {
    fprintf(stderr,"\nSysDevEnum: CoCreateInstance\n");
    return NULL;
  }

  hr = pSysDevEnum->lpVtbl->CreateClassEnumerator(pSysDevEnum,
    category,&pEnumCat,0);
  if (hr != S_OK) {
    pEnumCat->lpVtbl->Release(pEnumCat);
    fprintf(stderr,"\nSysDevEnum: CreateClassEnumerator\n");
    return NULL;
  }

  VariantInit(&varName);
  while(pEnumCat->lpVtbl->Next(pEnumCat,1,&pMoniker,&cFetched) == S_OK) {

    hr = pMoniker->lpVtbl->BindToStorage(pMoniker,
      0,0,&IID_IPropertyBag,(void **)&pPropBag);
    if (hr != S_OK) {
      pMoniker->lpVtbl->Release(pMoniker);
      break;
    }
    hr = pPropBag->lpVtbl->Read(pPropBag,L"FriendlyName",&varName,0);
    pPropBag->lpVtbl->Release(pPropBag);

    if (hr != S_OK || wcscmp(varName.bstrVal,varFriendlyName) != 0) {
      pMoniker->lpVtbl->Release(pMoniker);
      continue;
    }

    hr = pMoniker->lpVtbl->BindToObject(pMoniker,
      NULL,NULL,&IID_IBaseFilter,(void**)&pFilter);
    pMoniker->lpVtbl->Release(pMoniker);
    if (hr == S_OK) break;
  }

  VariantClear(&varName);
  pEnumCat->lpVtbl->Release(pEnumCat);
  pSysDevEnum->lpVtbl->Release(pSysDevEnum);
  return pFilter;
}

static const GUID CLSID_KSCATEGORY_BDA_NETWORK_TUNER
 = {0x71985f48,0x1ca1,0x11d3,{0x9c,0xc8,0x00,0xc0,0x4f,0x79,0x71,0xe0}};
static const GUID CLSID_PropSet
 = {0x9E1781E1,0x9CB1,0x4407,{0xBB,0xCE,0x54,0x26,0xC8,0xD0,0x0A,0x4B}};


static HRESULT hduscard_set(
  IKsPropertySet *pIKsPropertySet,
  DWORD dwPropID,
  LPVOID pInstanceData,
  DWORD cbInstanceData,
  LPVOID pPropData,
  DWORD cbPropData)
{
  HRESULT hr;
  static BOOL sent = FALSE;
  DWORD timeout;
#ifdef _DEBUG
  DWORD t;
#endif

#ifdef _DEBUG
  t = GetTickCount();
#endif
  timeout = GetTickCount()+3000;
  do {
    hr = pIKsPropertySet->lpVtbl->Set(pIKsPropertySet,
      &CLSID_PropSet,dwPropID,
      pInstanceData,cbInstanceData,
      pPropData,cbPropData);
    if (sent)
      return hr;
    if (hr == 0x8007001f)
      Sleep(50);
  } while (hr == 0x8007001f && timeout > GetTickCount());

#ifdef _DEBUG
  t = GetTickCount()-t;
  if (t > 100)
    fprintf(stderr,"\nhduscard_send: time elapsed=%d.%03d\n",t/1000,t%1000);
#endif
  sent = TRUE;
  return hr;
}

static HRESULT hduscard_send(hdus_tuner tuner,
  BYTE *pbSendBuffer,DWORD cbSendLength)
{
  HRESULT hr;
  IKsPropertySet *pIKsPropertySet;
  BYTE buf[0x42];

  if (tuner == NULL) {
    fprintf(stderr,"\nhduscard_send: tuner == NULL\n");
    return S_FALSE;
  }
  if (cbSendLength > 0x3c) {
    fprintf(stderr,"\nhduscard_send: data too long.\n");
    return S_FALSE;
  }

  memset(buf,0,sizeof(buf));
  buf[1] = (BYTE)cbSendLength;
  memcpy(buf+2,pbSendBuffer,cbSendLength);

  hr = tuner->lpVtbl->QueryInterface(tuner,
    &IID_IKsPropertySet,(LPVOID*)&pIKsPropertySet);
  if (hr != S_OK) {
    fprintf(stderr,"\nhduscard_send: QueryInterface\n");
    return hr;
  }

  hr = hduscard_set(pIKsPropertySet,
    0x10,NULL,0,buf,0x42);
  pIKsPropertySet->lpVtbl->Release(pIKsPropertySet);
  if (hr != S_OK) {
    fprintf(stderr,"\nhduscard_send: IKsPropertySet.Set\nhr=%08X\ncmd=%02X\n",hr,pbSendBuffer[1]);
    return hr;
  }

  return S_OK;
}

static HRESULT hduscard_receive(hdus_tuner tuner,
  BYTE *pbRecvBuffer,DWORD *pcbRecvLength)
{
  HRESULT hr;
  IKsPropertySet *pIKsPropertySet;
  BYTE buf[0x42];
  DWORD m,timeout;
#ifdef _DEBUG
  DWORD t;
#endif

  if (tuner == NULL) {
    fprintf(stderr,"\nhduscard_receive: tuner == NULL\n");
    return S_FALSE;
  }
  if (*pcbRecvLength < 0x3c) {
    fprintf(stderr,"\nhduscard_receive: buffer too small.\n");
    return S_FALSE;
  }

  hr = tuner->lpVtbl->QueryInterface(tuner,
    &IID_IKsPropertySet,(LPVOID*)&pIKsPropertySet);
  if (hr != S_OK) {
    fprintf(stderr,"\nhduscard_receive: QueryInterface\n");
    return hr;
  }

#ifdef _DEBUG
  t = GetTickCount();
#endif
  timeout = GetTickCount()+3000;
  do {
    hr = pIKsPropertySet->lpVtbl->Get(pIKsPropertySet,
      &CLSID_PropSet,0x11,NULL,0,buf,0x42,&m);
    if (hr != S_OK) {
      pIKsPropertySet->lpVtbl->Release(pIKsPropertySet);
      fprintf(stderr,"\nhduscard_receive: IKsPropertySet.Get\n");
      return hr;
    }
    if (buf[0] != 1)
      Sleep(50);
  } while (buf[0] > 1 && timeout > GetTickCount());

#ifdef _DEBUG
  t = GetTickCount()-t;
  if (t > 100)
    fprintf(stderr,"\nhduscard_receive: time elapsed=%d.%03d\n",t/1000,t%1000);
#endif

  pIKsPropertySet->lpVtbl->Release(pIKsPropertySet);
  if (buf[0] != 1) {
    fprintf(stderr,"\nhduscard_receive: failed.\n");
    return S_FALSE;
  }

  if (*pcbRecvLength < buf[1]) {
    fprintf(stderr,"\nhduscard_receive: buffer too small(2)\n");
    return S_FALSE;
  }
  *pcbRecvLength = buf[1];
  memcpy(pbRecvBuffer,buf+2,buf[1]);

  return S_OK;
}

hdus_tuner hduscard_create(void)
{
  hdus_tuner tuner;

  /*CoInitialize(NULL);*/

  tuner = find_dev(&CLSID_KSCATEGORY_BDA_NETWORK_TUNER,
    L"SKNET HDTV BDA Digital Tuner_0");
  if (tuner == NULL) {
    fprintf(stderr,"\nhduscard_create: find_dev\n");
    return NULL;
  }
  return tuner;
}

void hduscard_release(hdus_tuner tuner)
{
  if (tuner == NULL)
    return;
  tuner->lpVtbl->Release(tuner);
#if 0
  CoUninitialize();
#endif
}

LONG hduscard_transmit(hdus_tuner tuner,
  BYTE *pbSendBuffer,
  DWORD cbSendLength,
  BYTE *pbRecvBuffer,
  DWORD *pcbRecvLength
)
{
  if (tuner == NULL) {
    fprintf(stderr,"\nhdus_transmit: tuner == NULL\n");
    return SCARD_F_INTERNAL_ERROR;
  }
  if (hduscard_send(tuner,pbSendBuffer,cbSendLength) != S_OK)
    return SCARD_F_INTERNAL_ERROR;
  Sleep(50);
  if (hduscard_receive(tuner,pbRecvBuffer,pcbRecvLength) != S_OK)
   return SCARD_F_INTERNAL_ERROR;
  return SCARD_S_SUCCESS;
}
